MODULE DDDetectors

!XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
! DDDetectors
!    Routines for initializing and modifying the detector setup,
!    notatably by setting isotopes to use, loading efficiencies from
!    file, and tabulating form factors.
!XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

USE DDTypes
USE DDCommandLine
USE DDNuclear
USE DDUtils
USE DDInput
USE DDOutput

IMPLICIT NONE

PUBLIC :: DDCalc_GetDetector,DDCalc_SetDetector,&
          DDCalc_InitDetectorDataCommandLine
INTERFACE DDCalc_GetDetector
  MODULE PROCEDURE GetDetector
END INTERFACE
INTERFACE DDCalc_SetDetector
  MODULE PROCEDURE SetDetector
END INTERFACE
INTERFACE DDCalc_InitDetectorDataCommandLine
  MODULE PROCEDURE InitDetectorDataCommandLine
END INTERFACE

CONTAINS


!=======================================================================
! DETECTOR EFFICIENCY/RESPONSE ROUTINES
!=======================================================================

! ----------------------------------------------------------------------
! Sets the efficiencies to those found in the given file, presumably
! an efficiency or efficiencies file generated by TPCMC.
! 
! Required input argument:
!     file       The file to load efficiencies from.
! Required output arguments:
!     NE         Number of recoil energy tabulation points loaded
!     E          Allocatable array to contain energies [keV].
!                Allocated to size [1:NE].
!     Neff       Number of interval/bin efficiencies loaded.
!     eff        Allocatable array to contain efficiency curves for
!                the total range and each interval/bin.  Allocated to
!                size [1:NE,0:Neff].
! 
SUBROUTINE LoadEfficiencyFile(file,NE,E,Neff,eff)
  IMPLICIT NONE
  CHARACTER(LEN=*), INTENT(IN) :: file
  INTEGER, INTENT(OUT) :: NE,Neff
  REAL*8, ALLOCATABLE, INTENT(OUT) :: E(:),eff(:,:)
  LOGICAL :: status
  INTEGER :: Kcol,Keff,Nrow,Ncol,Nvalid
  REAL*8, ALLOCATABLE :: data(:,:)
  
  ! Load table from file
  CALL LoadTable(file=file,Nrow=Nrow,Ncol=Ncol,data=data,status=status)
  IF ((.NOT. status) .OR. (Ncol .LT. 2)) THEN
    WRITE(0,*) 'ERROR: Failed to load data from file ' // TRIM(file) // '.'
    STOP
  END IF
  
  ! Energies
  NE = Nrow
  ALLOCATE(E(NE))
  E = data(:,1)
  
  ! Find number of valid efficiency columns
  Nvalid = 0
  DO Kcol = 2,Ncol
    IF (ALL(data(:,Kcol) .GE. 0d0) .AND. ALL(data(:,Kcol) .LE. 1.00001d0)) Nvalid = Nvalid + 1
  END DO
  IF (Nvalid .LE. 0) THEN
    WRITE(0,*) 'ERROR: Failed to find valid data in file ' // TRIM(file) // '.'
    STOP
  END IF
  
  ! Now get efficiencies
  Neff = Nvalid - 1
  ALLOCATE(eff(NE,0:Neff))
  Keff = 0
  DO Kcol = 2,Ncol
    IF (ALL(data(:,Kcol) .GE. 0d0) .AND. ALL(data(:,Kcol) .LE. 1.00001d0) &
        .AND. (Keff .LE. Neff)) THEN
      eff(:,Keff) = data(:,Kcol)
      Keff = Keff + 1
    END IF
  END DO
  
END SUBROUTINE


! ----------------------------------------------------------------------
! FOR FUTURE USE
! Retrieves the efficiencies found in the given file, presumably an
! efficiency or efficiencies file generated by TPCMC.
! 
! Required input argument:
!     file       The file to load efficiencies from.
! Required output argument:
!     effS       DetectorEfficiencyStruct to contain tabulated
!                efficiency data.
! 
SUBROUTINE NEWLoadEfficiencyFile(file,effS)
  IMPLICIT NONE
  CHARACTER(LEN=*), INTENT(IN) :: file
  TYPE(DetectorEfficiencyStruct), INTENT(OUT) :: effS
  LOGICAL :: status
  INTEGER :: Kcol,Keff,Nrow,Ncol,Nvalid
  REAL*8, ALLOCATABLE :: data(:,:)
  
  effS%file = file
  
  ! Load table from file
  CALL LoadTable(file=file,Nrow=Nrow,Ncol=Ncol,data=data,status=status)
  IF ((.NOT. status) .OR. (Ncol .LT. 2)) THEN
    WRITE(0,*) 'ERROR: Failed to load data from file ' // TRIM(file) // '.'
    STOP
  END IF
  
  ! Energies
  effS%NE = Nrow
  ALLOCATE(effS%E(effS%NE))
  effS%E = data(:,1)
  
  ! Find number of valid efficiency columns
  Nvalid = 0
  DO Kcol = 2,Ncol
    IF (ALL(data(:,Kcol) .GE. 0d0) .AND. ALL(data(:,Kcol) .LE. 1.00001d0)) Nvalid = Nvalid + 1
  END DO
  IF (Nvalid .LE. 0) THEN
    WRITE(0,*) 'ERROR: Failed to find valid data in file ' // TRIM(file) // '.'
    STOP
  END IF
  
  ! Now get efficiencies
  effS%Neff = Nvalid
  ALLOCATE(effS%eff(effS%NE,0:effS%Neff))
  Keff = 0
  DO Kcol = 2,Ncol
    IF (ALL(data(:,Kcol) .GE. 0d0) .AND. ALL(data(:,Kcol) .LE. 1.00001d0) &
        .AND. (Keff .LE. effS%Neff)) THEN
      effS%eff(:,Keff) = data(:,Kcol)
      Keff = Keff + 1
    END IF
  END DO
  
END SUBROUTINE


! ----------------------------------------------------------------------
! Retabulates the given efficiency data at the given array of energies.
! The efficiency is taken to be zero beyond the tabulation range of
! the provided efficiency data.
! 
! Input arguments:
!     NEeff      Number of recoil energy tabulation points for raw
!                efficiency data.
!     Eeff       Array of size [1:NEeff] containing tabulation energies
!                for raw efficiency data [keV].
!     Neff       Number of interval/bin efficiencies in raw efficiency
!                data (0 if total only).
!     eff        Array of size [1:NEeff,0:Neff] containing raw
!                efficiency data.  The second index is over the
!                interval/bin number (0 for total range).
!     NE         Number of recoil energies in desired tabulation.
!     E          Array of size [1:NE] containing desired tabulation
!                energies [keV].
!     intervals  LOGICAL indicating if calculations for intervals/bins
!                will be performed (necessary for max gap).  If .FALSE.,
!                efficiency data for intervals/bins will be dropped.
! Output argument:
!     eff        Allocatable array to contain efficiency curves for
!                the total range and each interval/bin at the energies
!                in E.  Allocated to size [1:NE,0:Neff] or [1:NE,0:0],
!                depending on the intervals flag.
! 
SUBROUTINE RetabulateEfficiency(NEeff,Eeff,Neff,eff,NE,E,intervals,eff0)
  IMPLICIT NONE
  LOGICAL, INTENT(IN) :: intervals
  INTEGER, INTENT(IN) :: NEeff,Neff,NE
  REAL*8, INTENT(IN) :: Eeff(1:NEeff),eff(1:NEeff,0:Neff),E(1:NE)
  REAL*8, ALLOCATABLE, INTENT(OUT) :: eff0(:,:)
  INTEGER :: Neff0,Keff,K0,K0start
  REAL*8 :: f
  
  ! Allocate output array
  IF (intervals) THEN
    Neff0 = Neff
  ELSE
    Neff0 = 0
  END IF
  ALLOCATE(eff0(1:NE,0:Neff0))
  
  ! Special cases
  IF (E(NE) .LT. Eeff(1)) THEN
    eff0 = 0d0
    RETURN
  ELSE IF (E(1) .GT. Eeff(NEeff)) THEN
    eff0 = 0d0
    RETURN
  END IF
  
  ! Energies below tabulation range
  K0 = 1
  DO WHILE (E(K0) .LT. Eeff(1))
    eff0(K0,:) = 0d0
    K0 = K0 + 1
  END DO
  
  ! Cycle over remaining energies
  Keff    = -1
  K0start = K0
  DO K0 = K0start,NE
    Keff = BSearch(NEeff,Eeff,E(K0),Keff)
    ! Below tabulation range: set to zero (already accounted for)
    !IF (Keff .LE. 0) THEN
    !  eff0(K0,:) = 0d0
    ! Within tabulation range: linear interpolation
    IF (Keff .LT. NEeff) THEN
      ! Determine fractional distance between tabulation points
      IF (Eeff(Keff) .EQ. Eeff(Keff+1)) THEN
        f = 0.5d0
      ELSE
        f = (E(K0)-Eeff(Keff)) / (Eeff(Keff+1)-Eeff(Keff))
      END IF
      eff0(K0,:) = (1d0-f)*eff(Keff,0:Neff0) + f*eff(Keff+1,0:Neff0)
    ! Above tabulation range: set to zero
    ELSE
      IF (E(K0) .EQ. Eeff(NEeff)) THEN
        eff0(K0,:) = eff(NEeff,0:Neff0)
      ELSE
        eff0(K0,:) = 0d0
      END IF
      ! Can skip remaining points
      IF (K0 .LT. NE) THEN
        eff0(K0+1:NE,:) = 0d0
      END IF
      EXIT
    END IF
  END DO
  
END SUBROUTINE



!=======================================================================
! DETECTOR SETUP ROUTINES
!=======================================================================

! ----------------------------------------------------------------------
! Get various detector quantities.
! Note this only allows access to fixed quantities, not quantities that
! are recalculated for each WIMP (see GetRates() for that).
! 
! Optional input argument:
!   D           The DetectorStruct structure to extract detector
!               quantities from.  If not given, a default structure
!               (internally stored) will be used.
! Optional output arguments:
!   mass        Detector fiducial mass [kg]
!   time        Detector exposure time [day]
!   exposure    Detector exposure [kg day]
!   Nevents     Number of observed events
!   background  Average expected background events
!   Niso        Number of isotopes
!   Ziso        Allocatable integer array to be filled with isotope
!               atomic numbers. Allocated to size [1:Niso].
!   Aiso        Allocatable integer array to be filled with isotope
!               atomic mass numbers. Allocated to size [1:Niso].
!   fiso        Allocatable real array to be filled with isotope
!               mass fractions. Allocated to size [1:Niso].
!   Miso        Allocatable real array to be filled with isotope
!               nuclear masses [GeV]. Allocated to size [1:Niso].
!   NE          Number of tabulated recoil energies E
!   E           Allocatable real array to be filled with recoil energies
!               [keV]. Allocated to size [1:NE].
!   eff_file    File from which efficiencies were read
!   NEeff       Number of recoil energies E for efficiencies tabulation
!   Eeff        Allocatable real array to be filled with recoil energies
!               used for efficiencies tabulation [keV]. Allocated to
!               size [1:NEeff].
!   Neff        Number of subintervals for which efficiencies are
!               available (0 for only total interval)
!   eff         Allocatable dimension=2 array containing efficiencies
!               as a function of recoil energy. Allocated to size
!               [1:NEeff,0:Neff], where the first index is over recoil
!               energies and the second index is over the sub-interval
!               number (0 for the total interval).
!   Wsi,Wsd     Allocatable dimension=3 array containing weighted form
!               factors for spin-independent (SI) and spin-dependent
!               (SD) couplings.  Allocated to size [-1:1,1:NE,1:Niso].
!   intervals   LOGICAL indicating if rates for intervals/bins are
!               to be calculated (used for max gap).
! 
SUBROUTINE GetDetector(D,mass,time,exposure,Nevents,background,         &
                       Niso,Ziso,Aiso,fiso,Miso,NE,E,                   &
                       eff_file,NEeff,Eeff,Neff,eff,                    &
                       Wsi,Wsd,intervals)
  IMPLICIT NONE
  TYPE(DetectorStruct), INTENT(IN) :: D
  LOGICAL, INTENT(OUT), OPTIONAL :: intervals
  CHARACTER(LEN=*), INTENT(OUT), OPTIONAL :: eff_file
  INTEGER, INTENT(OUT), OPTIONAL :: Nevents,Niso,NE,NEeff,Neff
  INTEGER, ALLOCATABLE, INTENT(OUT), OPTIONAL :: Ziso(:),Aiso(:)
  REAL*8, INTENT(OUT), OPTIONAL :: mass,time,exposure,background
  REAL*8, ALLOCATABLE, INTENT(OUT), OPTIONAL :: fiso(:),Miso(:),E(:),   &
          Eeff(:),eff(:,:),Wsi(:,:,:),Wsd(:,:,:)
  
  ! Exposures
  IF (PRESENT(mass))     mass     = D%mass
  IF (PRESENT(time))     time     = D%time
  IF (PRESENT(exposure)) exposure = D%exposure
  
  ! Observed events and expected background events
  IF (PRESENT(Nevents))    Nevents    = D%Nevents
  IF (PRESENT(background)) background = D%MuBackground
  
  ! Isotope data
  IF (PRESENT(Niso)) Niso = D%Niso
  IF (PRESENT(Ziso)) THEN
    ALLOCATE(Ziso(D%Niso))
    Ziso = D%Ziso
  END IF
  IF (PRESENT(Aiso)) THEN
    ALLOCATE(Aiso(D%Niso))
    Aiso = D%Aiso
  END IF
  IF (PRESENT(fiso)) THEN
    ALLOCATE(fiso(D%Niso))
    fiso = D%fiso
  END IF
  IF (PRESENT(Miso)) THEN
    ALLOCATE(Miso(D%Niso))
    Miso = D%Miso
  END IF
  
  ! Recoil energies
  IF (PRESENT(NE)) NE = D%NE
  IF (PRESENT(E)) THEN
    ALLOCATE(E(D%NE))
    E = D%E
  END IF
  
  ! Efficiencies
  IF (PRESENT(eff_file)) eff_file = D%eff_file
  IF (PRESENT(NEeff))    NEeff    = D%NEeff
  IF (PRESENT(Eeff)) THEN
    ALLOCATE(E(D%NEeff))
    Eeff = D%Eeff
  END IF
  IF (PRESENT(Neff))     Neff     = D%Neff
  IF (PRESENT(eff)) THEN
    ALLOCATE(eff(D%NEeff,0:D%Neff))
    eff = D%eff
  END IF
  
  ! Weighted form factors
  IF (PRESENT(Wsi)) THEN
    ALLOCATE(Wsi(-1:1,D%NE,0:D%Niso))
    Wsi = D%Wsi
  END IF
  IF (PRESENT(Wsd)) THEN
    ALLOCATE(Wsd(-1:1,D%NE,0:D%Niso))
    Wsd = D%Wsd
  END IF
  
  ! Calculate rates for intervals/bins?
  IF (PRESENT(intervals)) intervals = D%intervals
  
END SUBROUTINE


! ----------------------------------------------------------------------
! Set various detector quantities.  Making changes will reset all
! per-WIMP calculations like rates.
! Note this only allows access to fixed quantities, or quantities that
! need be calculated only once, not quantities that are recalculated
! for each WIMP (see SetRates() for that).
! 
! Optional input/output argument:
!   D           The DetectorStruct structure containing detector
!               data to be modified.  If not given, a default structure
!               (internally stored) will be used.
! Optional input arguments:
!   mass        Detector fiducial mass [kg]
!   time        Detector exposure time [day]
!   exposure    Detector exposure [kg day]
!   Nevents     Number of observed events
!   background  Average expected background events
! Optional isotope-related input arguments.  Niso must be given for
! any of the arrays to be used, regardless if the number has changed.
! If Niso changes, then all of the A/Z/f arrays must be specified for
! changes to take effect (otherwise these inputs are ignored).
!   Niso        Number of isotopes
!   Ziso        Integer array of size [1:Niso] containing atomic
!               numbers.
!   Aiso        Integer array of size [1:Niso] containing atomic
!               mass numbers.
!   fiso        Array of size [1:Niso] containing isotope mass
!               fractions.
! Pre-populated isotopic abundances for compounds with given
! stoichiometry.  The first two of these must be given to take effect
! (otherwise these inputs are ignored).
!   Nelem       Number of elements in compound.
!   Zelem       Integer array of size [1:Nelem] containing atomic
!               numbers of compound elements.
!   stoich      Integer array of size [1:Nelem] containing compound
!               stoichiometry.  For example, CF3Cl would be {1,3,1}.
!               Default is 1 for each element.
! Optional recoil energy tabulation arguments.  Both are required for
! changes to take effect.
!   NE          Number of tabulated recoil energies E
!   E           Array of size [1:NE] containing recoil energies [keV].
!               This defines the energies used for calculating dR/dE
!               and its integral.
! Optional efficiency curve(s) arguments.  Efficiencies can be provided
! by file or by providing tabulation data.  If tabulation data is given,
! all of NEeff, Eeff, Neff, and eff must be provided.
!   eff_file    File from which efficiencies shoud be read.
!   eff_filename Sets the stored file name _without_ loading any data
!               from the file.
!   NEeff       Number of recoil energy tabulation points for
!               efficiency data.
!   Eeff        Array of size [1:NEeff] containing tabulation energies
!               for efficiency data [keV].
!   Neff        Number of sub-interval/bin efficiencies in efficiency
!               data (0 if total only).
!   eff         Array of size [1:NEeff,0:Neff] containing efficiencies
!               as a function of recoil energy, where the first index
!               is over recoil energies and the second index is over
!               the sub-interval number (0 for the total interval).
! Optional analysis-related arguments.
!   intervals   Specify if rates for sub-intervals should be calculated;
!               otherwise, only the full interval is used for
!               calculations (default: true).  This basically specifies
!               if maximum gap or a Poisson is used in exclusion
!               constraints.
!   Emin        If given, sets all efficiencies below the given energy
!               [keV] to zero, removing all contributions from recoils
!               at lower energies.
! Optional form factor arguments.  The NE and Niso arguments must also
! be given for changes to take effect.
!   Wsi,Wsd     Arrays of size [-1:1,1:NE,1:Niso] containing weighted
!               form factors for spin-independent (SI) and spin-
!               dependent (SD) couplings.  Note: these are replaced with
!               internal calculations if the energy tabulation changes
!               in subsequent calls.  Should not be used in combination
!               with Emin.
! 
SUBROUTINE SetDetector(D,mass,time,exposure,Nevents,background,         &
                       Niso,Ziso,Aiso,fiso,Nelem,Zelem,stoich,          &
                       NE,E,                                            &
                       eff_file,eff_filename,NEeff,Eeff,Neff,eff,       &
                       intervals,Emin,Wsi,Wsd)
  IMPLICIT NONE
  TYPE(DetectorStruct), INTENT(INOUT) :: D
  CHARACTER(LEN=*), INTENT(IN), OPTIONAL :: eff_file,eff_filename
  LOGICAL, INTENT(IN), OPTIONAL :: intervals
  INTEGER, INTENT(IN), OPTIONAL :: Nevents,Niso,Nelem,NE,NEeff,Neff
  !INTEGER, INTENT(IN), OPTIONAL :: Ziso(Niso),Aiso(Niso),Zelem(Nelem), &
  !                                 stoich(Nstoich)
  INTEGER, INTENT(IN), OPTIONAL :: Ziso(:),Aiso(:),Zelem(:),stoich(:)
  REAL*8, INTENT(IN), OPTIONAL :: mass,time,exposure,background,Emin
  !REAL*8, INTENT(IN), OPTIONAL :: fiso(Niso),E(NE),eff(NE,0:Neff),      &
  !        Wsi(-1:1,NE,Niso),Wsd(-1:1,NE,Niso)
  REAL*8, INTENT(IN), OPTIONAL :: fiso(:),E(:),Eeff(:),eff(:,0:),       &
          Wsi(-1:,:,:),Wsd(-1:,:,:)
  LOGICAL :: iso_change,E_change,eff_change
  INTEGER :: KE,Kiso,Neff0
  INTEGER, ALLOCATABLE :: stoich0(:)
  REAL*8 :: f,dx
  ! For setting default energy tabulation (logarithmic spacing)
  INTEGER, PARAMETER :: E_PER_DECADE = 100
  REAL*8, PARAMETER :: DEFAULT_EMIN = 0.1d0
  REAL*8, PARAMETER :: DEFAULT_EMAX = 1000d0
  
  ! Indicate if quantities changed, indicating need for
  ! array resizing and initialization
  iso_change = .FALSE.
  E_change   = .FALSE.
  eff_change = .FALSE.
  
  ! Exposures
  IF (PRESENT(mass)) THEN
    D%mass     = mass
    D%exposure = D%mass * D%time
  END IF
  IF (PRESENT(time)) THEN
    D%time     = time
    D%exposure = D%mass * D%time
  END IF
  IF (PRESENT(exposure)) THEN
    D%mass     = -1d0
    D%time     = -1d0
    D%exposure = exposure
  END IF
  
  ! Observed events and expected background events
  IF (PRESENT(Nevents))    D%Nevents      = Nevents
  IF (PRESENT(background)) D%MuBackground = background
  
  ! Isotope data
  IF (PRESENT(Niso)) THEN
    IF (Niso .EQ. D%Niso) THEN
      IF (PRESENT(Ziso)) THEN
        D%Ziso = Ziso(1:Niso)
        iso_change = .TRUE.
      END IF
      IF (PRESENT(Aiso)) THEN
        D%Aiso = Aiso(1:Niso)
        iso_change = .TRUE.
      END IF
      IF (PRESENT(fiso)) THEN
        D%fiso = fiso(1:Niso)
        iso_change = .TRUE.
      END IF
      IF (iso_change) THEN
        D%Miso = IsotopeMass(D%Ziso,D%Aiso)
      END IF
    ELSE IF (PRESENT(Ziso) .AND. PRESENT(Aiso) .AND. PRESENT(fiso)) THEN
      IF (ALLOCATED(D%Ziso)) DEALLOCATE(D%Ziso)
      IF (ALLOCATED(D%Aiso)) DEALLOCATE(D%Aiso)
      IF (ALLOCATED(D%fiso)) DEALLOCATE(D%fiso)
      IF (ALLOCATED(D%Miso)) DEALLOCATE(D%Miso)
      ALLOCATE(D%Ziso(Niso),D%Aiso(Niso),D%fiso(Niso),D%Miso(Niso))
      D%Niso = Niso
      D%Ziso = Ziso(1:Niso)
      D%Aiso = Aiso(1:Niso)
      D%fiso = fiso(1:Niso)
      D%Miso = IsotopeMass(D%Ziso,D%Aiso)
      iso_change = .TRUE.
    END IF
  END IF
  IF (PRESENT(Nelem) .AND. PRESENT(Zelem)) THEN
    ALLOCATE(stoich0(Nelem))
    IF (PRESENT(stoich)) THEN
      stoich0 = stoich
    ELSE
      stoich0 = 1
    END IF
    CALL CompoundIsotopeList(Nelem,Zelem,stoich0,                       &
                             D%Niso,D%Ziso,D%Aiso,D%fiso,D%Miso)
    iso_change = .TRUE.
  END IF
  
  ! Energies
  IF (PRESENT(NE) .AND. PRESENT(E)) THEN
    IF (ALLOCATED(D%E) .AND. (NE .NE. D%NE)) DEALLOCATE(D%E)
    IF (.NOT. ALLOCATED(D%E)) ALLOCATE(D%E(NE))
    D%NE = NE
    D%E = E(1:NE)
    IF (ALLOCATED(D%E_cache)) DEALLOCATE(D%E_cache)
    ALLOCATE(D%E_cache(D%NE))
    D%E_cache = D%E
    E_change = .TRUE.
  END IF
  
  ! Set default energies if not initialized or NE <= 0.
  ! Will use logarithmic spacing, but including E=0.
  IF (PRESENT(NE)) THEN
    IF (NE .LE. 0) D%NE = NE
  END IF
  IF (D%NE .LE. 0) THEN
    IF (ALLOCATED(D%E)) DEALLOCATE(D%E)
    D%NE = NINT(E_PER_DECADE*(LOG10(DEFAULT_EMAX/DEFAULT_EMIN))) + 2
    ALLOCATE(D%E(D%NE))
    dx = 1d0/E_PER_DECADE
    D%E(1) = 0d0
    D%E(2) = DEFAULT_EMIN
    DO KE = 2, D%NE-1
      D%E(KE) = DEFAULT_EMIN * 10d0**((KE-2)*dx)
    END DO
    D%E(D%NE) = DEFAULT_EMAX
    IF (ALLOCATED(D%E_cache)) DEALLOCATE(D%E_cache)
    ALLOCATE(D%E_cache(D%NE))
    D%E_cache = D%E
    E_change = .TRUE.
  END IF
  
  ! Set efficiencies
  ! ...from file
  IF (PRESENT(eff_file)) THEN
    IF (eff_file .NE. '') THEN
      D%eff_file = eff_file
      CALL LoadEfficiencyFile(eff_file,D%NEeff,D%Eeff,D%Neff,D%eff)
      eff_change = .TRUE.
    END IF
  ! ...by arguments
  ELSE IF (PRESENT(NEeff) .AND. PRESENT(Eeff) .AND. PRESENT(Neff)       &
           .AND. PRESENT(eff)) THEN
    IF (ALLOCATED(D%Eeff)) DEALLOCATE(D%Eeff)
    IF (ALLOCATED(D%eff))  DEALLOCATE(D%eff)
    ALLOCATE(D%Eeff(NEeff),D%eff(NEeff,0:Neff))
    D%NEeff = NEeff
    D%Eeff  = Eeff(1:NEeff)
    D%Neff  = Neff
    D%eff   = eff(1:NEeff,0:Neff)
    eff_change = .TRUE.
  END IF
  
  ! Set efficiency to 100% if not initialized or NEeff <= 0.
  ! No intervals.
  IF (PRESENT(NEeff)) THEN
    IF (NEeff .LE. 0) D%NEeff = NEeff
  END IF
  IF (D%NEeff .LE. 0) THEN
    IF (ALLOCATED(D%Eeff)) DEALLOCATE(D%Eeff)
    IF (ALLOCATED(D%eff))  DEALLOCATE(D%eff)
    ALLOCATE(D%Eeff(2),D%eff(2,0:0))
    D%NEeff = 2
    D%Eeff  = (/ 0d0, HUGE(D%Eeff) /)
    D%Neff  = 0
    D%eff   = 1d0
    eff_change = .TRUE.
  END IF
  
  ! Saved efficiency file name
  IF (PRESENT(eff_filename)) D%eff_file = eff_filename
  
  ! Include sub-intervals?
  IF (PRESENT(intervals)) THEN
    IF (intervals .NEQV. D%intervals) eff_change = .TRUE.
    D%intervals = intervals
  END IF
  
  ! Apply threshold cut.
  ! We move all E < Emin tabulation points to Emin.
  IF (PRESENT(Emin)) THEN
    ! First reset to original tabulation
    D%E = MAX(D%E_cache,Emin)
    E_change = .TRUE.
  END IF
  
  ! Update efficiencies
  IF (eff_change .OR. E_change) THEN
    CALL RetabulateEfficiency(D%NEeff,D%Eeff,D%Neff,D%eff,          &
                              D%NE,D%E,D%intervals,D%eff0)
  END IF
  
  ! Weighted form factors (SI)
  IF (PRESENT(NE) .AND. PRESENT(Niso) .AND. PRESENT(Wsi)) THEN
    IF (ALLOCATED(D%Wsi)) DEALLOCATE(D%Wsi)
    ALLOCATE(D%Wsi(-1:1,NE,Niso))
    D%Wsi = Wsi(-1:1,1:NE,1:Niso)
  ELSE IF ((iso_change .OR. E_change) .AND. (D%Niso .GE. 0)            &
           .AND. (D%NE .GE. 0)) THEN
    IF (ALLOCATED(D%Wsi)) DEALLOCATE(D%Wsi)
    ALLOCATE(D%Wsi(-1:1,D%NE,D%Niso))
    DO Kiso = 1,D%Niso
      CALL CalcWSI(D%Ziso(Kiso),D%Aiso(Kiso),D%NE,                   &
                   EToQ(D%E,D%Miso(Kiso)),D%Wsi(:,:,Kiso))
    END DO
  END IF
  
  ! Weighted form factors (SD)
  IF (PRESENT(NE) .AND. PRESENT(Niso) .AND. PRESENT(Wsd)) THEN
    IF (ALLOCATED(D%Wsd)) DEALLOCATE(D%Wsd)
    ALLOCATE(D%Wsd(-1:1,NE,Niso))
    D%Wsd = Wsd(-1:1,1:NE,1:Niso)
  ELSE IF ((iso_change .OR. E_change) .AND. (D%Niso .GE. 0)            &
           .AND. (D%NE .GE. 0)) THEN
    IF (ALLOCATED(D%Wsd)) DEALLOCATE(D%Wsd)
    ALLOCATE(D%Wsd(-1:1,D%NE,D%Niso))
    DO Kiso = 1,D%Niso
      CALL CalcWSD(D%Ziso(Kiso),D%Aiso(Kiso),D%NE,                   &
                   EToQ(D%E,D%Miso(Kiso)),D%Wsd(:,:,Kiso))
    END DO
  END IF
  
  ! Resize halo velocity arrays if necessary
  IF ((iso_change .OR. E_change) .AND. (D%Niso .GE. 0)                 &
      .AND. (D%NE .GE. 0)) THEN
    IF (ALLOCATED(D%vmin)) DEALLOCATE(D%vmin)
    ALLOCATE(D%vmin(D%NE,D%Niso))
    IF (ALLOCATED(D%eta))  DEALLOCATE(D%eta)
    ALLOCATE(D%eta(D%NE,D%Niso))
  END IF
  
  ! Number of intervals/bins to do calculations for.
  ! Used for array sizing below.
  IF (D%intervals) THEN
    Neff0 = D%Neff
  ELSE
    Neff0 = 0
  END IF
  
  ! Resize rate arrays if necessary
  IF ((E_change .OR. eff_change) .AND. (D%NE .GE. 0)                   &
      .AND. (D%Neff .GE. 0)) THEN
    IF (ALLOCATED(D%dRdEsi0)) DEALLOCATE(D%dRdEsi0)
    ALLOCATE(D%dRdEsi0(-1:1,D%NE))
    IF (ALLOCATED(D%dRdEsd0)) DEALLOCATE(D%dRdEsd0)
    ALLOCATE(D%dRdEsd0(-1:1,D%NE))
    IF (ALLOCATED(D%Rsi0)) DEALLOCATE(D%Rsi0)
    ALLOCATE(D%Rsi0(-1:1,0:Neff0))
    IF (ALLOCATED(D%Rsd0)) DEALLOCATE(D%Rsd0)
    ALLOCATE(D%Rsd0(-1:1,0:Neff0))
    IF (ALLOCATED(D%dRdEsi)) DEALLOCATE(D%dRdEsi)
    ALLOCATE(D%dRdEsi(D%NE))
    IF (ALLOCATED(D%dRdEsd)) DEALLOCATE(D%dRdEsd)
    ALLOCATE(D%dRdEsd(D%NE))
    IF (ALLOCATED(D%dRdE)) DEALLOCATE(D%dRdE)
    ALLOCATE(D%dRdE(D%NE))
    IF (ALLOCATED(D%Rsi)) DEALLOCATE(D%Rsi)
    ALLOCATE(D%Rsi(0:Neff0))
    IF (ALLOCATED(D%Rsd)) DEALLOCATE(D%Rsd)
    ALLOCATE(D%Rsd(0:Neff0))
    IF (ALLOCATED(D%R)) DEALLOCATE(D%R)
    ALLOCATE(D%R(0:Neff0))
  END IF
  
  ! Resize event arrays if necessary
  IF (eff_change .AND. (D%Neff .GE. 0)) THEN
    IF (ALLOCATED(D%MuSignalSI0)) DEALLOCATE(D%MuSignalSI0)
    ALLOCATE(D%MuSignalSI0(-1:1,0:Neff0))
    IF (ALLOCATED(D%MuSignalSD0)) DEALLOCATE(D%MuSignalSD0)
    ALLOCATE(D%MuSignalSD0(-1:1,0:Neff0))
    IF (ALLOCATED(D%MuSignalSI)) DEALLOCATE(D%MuSignalSI)
    ALLOCATE(D%MuSignalSI(0:Neff0))
    IF (ALLOCATED(D%MuSignalSD)) DEALLOCATE(D%MuSignalSD)
    ALLOCATE(D%MuSignalSD(0:Neff0))
    IF (ALLOCATED(D%MuSignal)) DEALLOCATE(D%MuSignal)
    ALLOCATE(D%MuSignal(0:Neff0))
  END IF
  
  ! Set all calculable quantities to zero
  IF ((iso_change .OR. E_change .OR. eff_change)                        &
       .AND. (D%Niso .GE. 0) .AND. (D%NE .GE. 0)                      &
       .AND. (D%Neff .GE. 0)) THEN
    D%vmin        = 0d0
    D%eta         = 0d0
    D%dRdEsi0     = 0d0
    D%dRdEsd0     = 0d0
    D%Rsi0        = 0d0
    D%Rsd0        = 0d0
    D%dRdEsi      = 0d0
    D%dRdEsd      = 0d0
    D%dRdE        = 0d0
    D%Rsi         = 0d0
    D%Rsd         = 0d0
    D%R           = 0d0
    D%MuSignalSI0 = 0d0
    D%MuSignalSD0 = 0d0
    D%MuSignalSI  = 0d0
    D%MuSignalSD  = 0d0
    D%MuSignal    = 0d0
  END IF
  
END SUBROUTINE


!-----------------------------------------------------------------------
! Initializes the detector setup from command-line parameters.
! 
! Possible options:
!   --mass=<value>       ! Detector fiducial mass [kg]
!   --time=<value>       ! Detector exposure time [day]
!   --exposure=<value>   ! Detector exposure [kg day]
!   --events=<N>         ! Number of observed events
!   --background=<value> ! Average expected background events
! Isotope-related options specifying the atomic number, atomic mass
! number, and mass fraction of the isotopes.  Must specify all three
! and they must have the same number of isotopes for changes to take
! effect:
!   --isotope-Z=<Z1>,<Z2>,<Z3>,...
!   --isotope-A=<A1>,<A2>,<A3>,...
!   --isotope-f=<f1>,<f2>,<f3>,...
! Use naturally occurring isotopes for the elements in the given
! compound specified by atomic numbers and the stoichiometry.  If
! stoichiometry is not given, it is assumed to be 1 for each element
! (equal abundances):
!   --element-Z=<Z1>,<Z2>,<Z3>,...
!   --stoichiometry=<stoich1>,<stoich2>,<stoich3>,...
! Shortcuts for specific materials by name (using natural abundances):
!   --argon
!   --germanium
!   --sodium-iodide
!   --silicon
!   --xenon
! Efficiency related options:
!   --file=<file>        ! File from which tabulated efficiencies should be read.
!                        ! First column is E [keV], any following consecutive
!                        ! columns containing values above 1.0 are ignored, and
!                        ! the remaining columns are taken to be efficiency curves.
!                        ! First efficiency is for full range, remaining efficiencies
!                        ! are for sub-intervals.  If not given, a default is used.
!   --no-intervals       ! Disable use of sub-intervals, even if available in file.
!   --Emin=<value>       ! If given, sets all efficiencies below the given energy
!                        ! [keV] to zero, removing all contributions from recoils
!                        ! at lower energies.  Not reversible.
! 
SUBROUTINE InitDetectorDataCommandLine(Detector)

  IMPLICIT NONE
  TYPE(DetectorStruct), INTENT(INOUT) :: Detector
  INTEGER :: Nevents,Niso,Nelem,N1,N2,N3
  INTEGER, ALLOCATABLE :: Ziso(:),Aiso(:),Zelem(:),stoich(:)
  REAL*8 :: mass,time,exposure,background,Emin
  REAL*8, ALLOCATABLE :: fiso(:)
  
  Niso = -1
    
  ! Update isotopes
  IF (GetLongArgIntegers('isotope-Z',Ziso,N1)                           &
      .AND. GetLongArgIntegers('isotope-A',Aiso,N2)                     &
      .AND. GetLongArgReals('isotope-f',fiso,N3)) THEN
    IF ((N1 .EQ. N2) .AND. (N1 .EQ. N3)) THEN
      Niso = N1
      CALL SetDetector(Detector,Niso=Niso,Ziso=Ziso,Aiso=Aiso,fiso=fiso)
    END IF
  ELSE IF (GetLongArgIntegers('element-Z',Zelem,Nelem)) THEN
    IF (GetLongArgIntegers('stoichiometry',stoich,N2)) THEN
      IF (N2 .EQ. Nelem) THEN
        CALL SetDetector(Detector,Nelem=Nelem,Zelem=Zelem,stoich=stoich)
      END IF
    ELSE
      IF (ALLOCATED(stoich)) DEALLOCATE(stoich)
      ALLOCATE(stoich(Nelem))
      stoich = 1
      CALL SetDetector(Detector,Nelem=Nelem,Zelem=Zelem,stoich=stoich)
    END IF
  ELSE IF (GetLongArg('argon')) THEN
    CALL SetDetector(Detector,Nelem=1,Zelem=(/18/))
  ELSE IF (GetLongArg('germanium')) THEN
    CALL SetDetector(Detector,Nelem=1,Zelem=(/32/))
  ELSE IF (GetLongArg('silicon')) THEN
    CALL SetDetector(Detector,Nelem=1,Zelem=(/14/))
  ELSE IF (GetLongArg('xenon')) THEN
    CALL SetDetector(Detector,Nelem=1,Zelem=(/54/))
  ELSE IF (GetLongArg('sodium-iodide')) THEN
    CALL SetDetector(Detector,Nelem=2,Zelem=(/11,53/),stoich=(/1,1/))
  END IF
  
  ! Update exposures
  IF (GetLongArgReal('mass',mass))         CALL SetDetector(Detector,mass=mass)
  IF (GetLongArgReal('time',time))         CALL SetDetector(Detector,time=time)
  IF (GetLongArgReal('exposure',exposure)) CALL SetDetector(Detector,exposure=exposure)
  
  ! Update observed events and expected background events
  IF (GetLongArgInteger('events',Nevents))     CALL SetDetector(Detector,Nevents=Nevents)
  IF (GetLongArgReal('background',background)) CALL SetDetector(Detector,background=background)
  
  ! Set an energy threshold
  IF (GetLongArgReal('Emin',Emin)) THEN
    CALL SetDetector(Detector,Emin=Emin)
  END IF
  
END SUBROUTINE

  
END MODULE
